"""Class to hold order type."""
from tradeo.order_operations import OrderOperations


class OrderType:
  """Class to hold order type."""

  def __init__(self, buy: bool, market: bool) -> None:
    """Order Type class for order_type attribute in the Order class.

    Args:
    - buy (bool): True if the order is a buy order,
    False if it is a sell order.
    - market (bool): True if the order is a market order,
    False if it is a pending order.
    """
    self.buy = buy
    self.sell = not buy
    self.market = market
    self.pending = not market

    self.__build()

  def __build(self) -> None:
    """Define the real order type.

    For non-market operations, "limit" type orders will be used by default,
    although they may automatically change to "stop" type orders when creating
    an order if the bid or ask value requires it.
    """
    if self.buy and self.market:
      self.value = OrderOperations.BUY
    elif self.buy and not self.market:
      self.value = OrderOperations.BUYLIMIT
    elif not self.buy and self.market:
      self.value = OrderOperations.SELL
    elif not self.buy and not self.market:
      self.value = OrderOperations.SELLLIMIT

  def __eq__(self, value: object) -> bool:
    """Check if the object is equal to another object."""
    return (
        isinstance(value, OrderType)
        and self.buy == value.buy
        and self.sell == value.sell
        and self.market == value.market
        and self.pending == value.pending
    )


def get_order_type_from_str(order_type_str: str) -> OrderType:
  """Get the order type object from a string."""
  if order_type_str == OrderOperations.BUYLIMIT:
    ot = OrderType(buy=True, market=False)
    ot.value = OrderOperations.BUYLIMIT
  elif order_type_str == OrderOperations.BUYSTOP:
    ot = OrderType(buy=True, market=False)
    ot.value = OrderOperations.BUYSTOP
  elif order_type_str == OrderOperations.SELLLIMIT:
    ot = OrderType(buy=False, market=False)
    ot.value = OrderOperations.SELLLIMIT
  elif order_type_str == OrderOperations.SELLSTOP:
    ot = OrderType(buy=False, market=False)
    ot.value = OrderOperations.SELLSTOP
  elif order_type_str == OrderOperations.BUY:
    ot = OrderType(buy=True, market=True)
    ot.value = OrderOperations.BUY
  elif order_type_str == OrderOperations.SELL:
    ot = OrderType(buy=False, market=True)
    ot.value = OrderOperations.SELL
  else:
    raise ValueError(f'Invalid order type: {order_type_str}')

  return ot



"""Class to hold order data."""
from tradeo.order_type import OrderType


class OrderPrice:
  """Class to hold order price."""

  def __init__(
      self,
      price: float = 0,
      stop_loss: float = 0,
      take_profit: float = 0
  ):
    """Initialize the attributes."""
    self.price = price
    self.stop_loss = stop_loss
    self.take_profit = take_profit


class MutableOrderDetails:
  """Class to hold order details."""

  def __init__(
      self,
      prices: OrderPrice,
      lots: float = 0.01,
      expiration: int = 0
  ):
    """Initialize the attributes."""
    self._prices = prices
    self.lots = lots
    self.expiration = expiration

  @property
  def price(self):
    """Get the entry price."""
    return self._prices.price

  @property
  def stop_loss(self):
    """Get the stop loss."""
    return self._prices.stop_loss

  @property
  def take_profit(self):
    """Get the take profit."""
    return self._prices.take_profit


class ImmutableOrderDetails:
  """Class to hold order metadata."""

  def __init__(
          self,
          symbol: str,
          order_type: OrderType,
          magic: str,
          comment: str,
  ):
    """Initialize the attributes."""
    self.symbol = symbol
    self.order_type = order_type
    self.magic = magic
    self.comment = comment


class Order:
  """Class to hold order data.

  symbol (str): Symbol for which an order should be opened.
  order_type (str): Order type. Can be one of:
      'buy', 'sell', 'buylimit', 'selllimit', 'buystop', 'sellstop'
  lots (float): Volume in lots
  price (float): Price of the (pending) order. Can be zero
      for market orders.
  stop_loss (float): SL as absolute price. Can be zero
      if the order should not have an SL.
  take_profit (float): TP as absolute price. Can be zero
      if the order should not have a TP.
  magic (int): Used to identify the order.
      Could be used to specify the timestamp as well.
  comment (str): Order comment
  expiration (int): Expiration time given as timestamp in seconds.
      Can be zero if the order should not have an expiration time.

  """

  def __init__(
      self,
      mutable_details: MutableOrderDetails,
      immutable_details: ImmutableOrderDetails,
      ticket: int = 0,
      pnl: float = 0
  ):
    """Initialize the attributes."""
    self._mutable_details = mutable_details
    self._immutable_details = immutable_details
    self._ticket = ticket  # it would not be available until the order be filled
    self._pnl = pnl  # it would not be available until the order be filled

  def __eq__(self, __value: object) -> bool:
    """Check if the order is equal to another order."""
    return isinstance(__value, Order) and self.ticket == __value.ticket

  def __str__(self) -> str:
    """Return a string representation of the order."""
    return f'{self.comment} {self.symbol} {self.price} {self.magic}'

  @property
  def symbol(self) -> str:
    """Get the symbol."""
    return self._immutable_details.symbol

  @property
  def order_type(self) -> OrderType:
    """Get the order type."""
    return self._immutable_details.order_type

  @property
  def lots(self) -> float:
    """Get the lots."""
    return self._mutable_details.lots

  @property
  def price(self) -> float:
    """Get the price."""
    return self._mutable_details.price

  @property
  def stop_loss(self) -> float:
    """Get the stop loss."""
    return self._mutable_details.stop_loss

  @property
  def take_profit(self) -> float:
    """Get the take profit."""
    return self._mutable_details.take_profit

  @property
  def magic(self) -> str:
    """Get the magic number."""
    return self._immutable_details.magic

  @property
  def comment(self) -> str:
    """Get the comment."""
    return self._immutable_details.comment

  @property
  def expiration(self) -> int:
    """Get the expiration."""
    return self._mutable_details.expiration

  @property
  def ticket(self) -> int:
    """Get the ticket."""
    return self._ticket

  @property
  def pnl(self) -> float:
    """Get the gain or loss."""
    return self._pnl

  def risk_benefit(self) -> float:
    """Get the risk benefit."""
    a = abs(self.take_profit - self.price)
    b = abs(self.price - self.stop_loss)
    return a / b if b > 0 else 0



"""Script to collect different utilities."""
from time import sleep
from datetime import datetime
import subprocess
import typing as ty
from os.path import join
from pytz import BaseTzInfo
from pytz.tzinfo import DstTzInfo, StaticTzInfo

from tradeo.config import Config
from tradeo.paths import bash_path, get_default_path
from tradeo.files import Files
from tradeo.files import write_file, try_read_file
from tradeo.log import log

timezone_type = ty.Union[DstTzInfo, BaseTzInfo, StaticTzInfo]


def string_to_date_utc(
    str_date: str,
    date_format: str = '%Y.%m.%d %H:%M',
    from_timezone: timezone_type = Config.utc_timezone
) -> datetime:
  """Convert a string to a datetime object in UTC timezone."""
  r = from_timezone.localize(datetime.strptime(str_date, date_format))
  return r.astimezone(Config.utc_timezone)


def reboot_mt():
  """Reboot MetaTrader."""
  log.warning('Rebooting MetaTrader ...')
  scripts_path = bash_path()
  subprocess.Popen(['/usr/bin/sh', join(scripts_path, 'stop-mt.sh')])
  sleep(3)
  subprocess.Popen(['/usr/bin/sh', join(scripts_path, 'launch-mt5.sh')])
  sleep(60)


def create_magic_number() -> str:
  """Create a magic number based on the current date and time."""
  return str(
      round(datetime.now(Config.utc_timezone).timestamp())
  ).replace('.', '')


def reset_consecutive_times_down() -> None:
  """Reset the consecutive times down file."""
  write_file(Files.CONSECUTIVE_TIMES_DOWN.value, '0')


def get_last_balance() -> float:
  """Get the balance of the account."""
  path = get_default_path()
  return float(try_read_file(path / Files.LAST_BALANCE.value))


def get_consecutive_times_down() -> int:
  """Get the consecutive times down."""
  path = get_default_path()
  return int(try_read_file(path / Files.CONSECUTIVE_TIMES_DOWN.value))


def increment_consecutive_times_down() -> None:
  """Increment the consecutive times down."""
  write_file(Files.CONSECUTIVE_TIMES_DOWN.value,
             str(get_consecutive_times_down() + 1))



"""OHLC class to encapsulate OHLC data."""
from pandas import DataFrame


class OHLC:
  """OHLC class."""

  def __init__(self, df: DataFrame):
    """Initialize the attributes."""
    self.open = df['open'].to_numpy()
    self.high = df['high'].to_numpy()
    self.low = df['low'].to_numpy()
    self.close = df['close'].to_numpy()

  def __len__(self):
    """Return the length of the OHLC data."""
    return len(self.open)



"""EMA Strategy."""
from typing import Union
from datetime import datetime

from tradeo.trading_methods import EMA, get_pivots, get_pip
from tradeo.ohlc import OHLC
from tradeo.utils import create_magic_number
from tradeo.strategies.strategy import Strategy
from tradeo.order import (Order,
                          MutableOrderDetails,
                          ImmutableOrderDetails,
                          OrderPrice,
                          OrderType)


class BasicStrategy(Strategy):
  """Strategy based on EMAs and price tendency."""

  def __init__(self):
    """Initialize the attributes."""
    super().__init__('BasicStrategy')

  def indicator(
      self,
      ohlc: OHLC,
      symbol: str,
      now_date: datetime
  ) -> Union[Order, None]:
    """Return an order if the strategy is triggered."""
    # This strategy does not use the date
    _ = now_date

    # Calculate the EMAs
    ema_20 = EMA(ohlc.close, 20)
    ema_50 = EMA(ohlc.close, 50)

    # Calculate the pivots
    p = get_pivots(ohlc.high, left=6, right=3, n_pivot=2, max_min='max')
    max_pivot_1 = p[0][0]  # most recent
    max_pivot_2 = p[1][0]
    p = get_pivots(ohlc.low, left=6, right=3, n_pivot=2, max_min='min')
    min_pivot_1 = p[0][0]  # most recent
    min_pivot_2 = p[1][0]

    # Check the conditions
    upper_tendency = max_pivot_1 > max_pivot_2 and min_pivot_1 > min_pivot_2
    lower_tendency = not upper_tendency and (
        max_pivot_1 < max_pivot_2) and (min_pivot_1 < min_pivot_2)

    # Obtain common features
    magic = create_magic_number()
    pip = get_pip(symbol)

    if ema_20[-1] >= ema_50[-1] and upper_tendency:
      return Order(
          MutableOrderDetails(
              OrderPrice(
                  take_profit=ema_20[-1] + 20 * pip,
                  stop_loss=ema_50[-1] - 10 * pip
              )),
          ImmutableOrderDetails(
              symbol=symbol, order_type=OrderType(buy=True, market=True),
              magic=magic, comment=self.strategy_name)
      )
    elif lower_tendency:
      return Order(
          MutableOrderDetails(
              OrderPrice(
                  take_profit=ema_20[-1] - 20 * pip,
                  stop_loss=ema_50[-1] + 10 * pip
              )),
          ImmutableOrderDetails(
              symbol=symbol, order_type=OrderType(buy=False, market=True),
              magic=magic, comment=self.strategy_name)
      )

    return None



"""This class blocks the execution writing a .block file."""
from tradeo.files import write_file, remove_file, file_exists


class Blocker:
  """Blocker class."""

  def __init__(self, name: str) -> None:
    """Initialize the Blocker."""
    self.name = f'{name}.block'

  def is_blocked(self) -> bool:
    """Check if the execution is blocked."""
    return file_exists(self.name)

  def __enter__(self) -> 'Blocker':
    """Blocker enter."""
    write_file(self.name)
    return self

  def __exit__(self, exc_type: str, exc_value: str, traceback: str) -> bool:
    """Blocker exit."""
    _, _, _ = exc_type, exc_value, traceback
    remove_file(self.name)
    return True



"""Common methods in trading strategies."""
from typing import List, Tuple
from pandas import Series
from numpy import ndarray

from tradeo.ohlc import OHLC
from tradeo.order_type import OrderType


def get_pip(symbol: str) -> float:
  """Return the pip value for symbol."""
  return 0.01 if 'JPY' in symbol else 0.0001


def get_pivots(
        data: ndarray,
        left: int,
        right: int,
        n_pivot: int,
        max_min: str
) -> List[Tuple[int, int]]:
  """Calculate the bar indices when a maximum or minimum occurs.

  left: number of greater/smaller bars to the left of the pivot
  right: number of greater/smaller bars to the right of the pivot
  n_pivot: number of pivots (tuples) to be found
  max_min: 'max' or 'min'

  Return [(value, index), (value, index), ..., n_pivot]
  """
  def getMax(data):
    m = max(data)
    i = data.index(m)
    return m, i

  def getMin(data):
    m = min(data)
    i = data.index(m)
    return m, i

  pivots = []  # [0] most recent
  data_list = list(data)
  size = left + right + 1
  a = -1 * (size + 1)
  b = -1
  d = data_list[a + 1:]
  m, i = getMax(d) if max_min == 'max' else getMin(d)
  if i == left and len(d) - 1 - i == right:
    pivots.append((m, len(d) - i - 1))
  while len(pivots) < n_pivot:
    d = data_list[a:b]
    if len(d) == size:
      m, i = getMax(d) if max_min == 'max' else getMin(d)
      if i == left and len(d) - 1 - i == right:
        pivots.append((m, -1 * b + right))
      a -= 1
      b -= 1
    else:
      break
  # Convert the indices to be prepared negatively
  # so that we can directly do for example: closes[-3]
  return [(i[0], -i[1] - 1) for i in pivots]


def EMA(
    data: ndarray,
    lookback: int,
    smoothing: int = 2
) -> List[float]:
  """Exponential Moving Average."""
  ema = [sum(data[:lookback]) / lookback]
  for price in data[lookback:]:
    ema.append(
        (price * (smoothing / (1 + lookback)))
        + ema[-1] * (1 - (smoothing / (1 + lookback)))
    )
  return ema


def RSI(data: ndarray, lookback: int) -> List[float]:
  """Relative Strength Index."""
  ret = Series(data).diff()
  up = []
  down = []
  for i in range(len(ret)):
    if ret[i] < 0:
      up.append(0)
      down.append(ret[i])
    else:
      up.append(ret[i])
      down.append(0)
  up_series = Series(up)
  down_series = Series(down).abs()
  up_ewm = up_series.ewm(com=lookback - 1, adjust=False).mean()
  down_ewm = down_series.ewm(com=lookback - 1, adjust=False).mean()
  rs = up_ewm / down_ewm
  return list(100 - (100 / (1 + rs)))


def SAR(  # noqa
  data: OHLC,
  af: float = 0.02,
  maxi: float = 0.2
) -> List[float]:
  """Calculate Simple Average Reversion."""
  high, low = data.high, data.low
  sig0, xpt0, af0 = True, high[0], af
  sar = [low[0] - (high - low).std()]
  for i in range(1, len(data)):
    sig1, xpt1, af1 = sig0, xpt0, af0
    l_min = min(low[i - 1], low[i])
    l_max = max(high[i - 1], high[i])
    if sig1:
      sig0 = low[i] > sar[-1]
      xpt0 = max(l_max, xpt1)
    else:
      sig0 = high[i] >= sar[-1]
      xpt0 = min(l_min, xpt1)
    if sig0 == sig1:
      sari = sar[-1] + (xpt1 - sar[-1]) * af1
      af0 = min(maxi, af1 + af)
      if sig0:
        af0 = af0 if xpt0 > xpt1 else af1
        sari = min(sari, l_min)
      else:
        af0 = af0 if xpt0 < xpt1 else af1
        sari = max(sari, l_max)
    else:
      af0 = af
      sari = xpt0
    sar.append(sari)
  return sar


def confirmation_pattern(data: OHLC, order_type: OrderType) -> bool:
  """Return True if different patterns are detected.

  https://s3.tradingview.com/z/zk1YPPgw_big.png
  https://www.tradingview.com/chart/EURUSD/zk1YPPgw-10-MOST-important-bar-patterns-to-profit-trade/
  """
  c1 = three_bar_reversal(data, order_type)
  c2 = pinbar_pattern(data, order_type)
  c3 = harami_pattern(data, order_type)
  return c1 or c2 or c3


def three_bar_reversal(data: OHLC, order_type: OrderType) -> bool:
  """Envelope Pattern.

  docs/images/buy_three_bar_reversal.png
  docs/images/sell_three_bar_reversal.png
  """
  opens = data.open
  highs = data.high
  lows = data.low
  closes = data.close
  if order_type.buy:
    return opens[-3] > closes[-3] and opens[-2] > closes[-2] and \
        opens[-1] < closes[-1] and lows[-3] > lows[-2] and \
        lows[-2] < lows[-1] and opens[-3] > highs[-2] and closes[-1] > highs[-2]
  else:
    return opens[-3] < closes[-3] and opens[-2] < closes[-2] and \
        opens[-1] > closes[-1] and highs[-3] < highs[-2] and \
        lows[-3] < lows[-2] and closes[-1] < lows[-2]


def pinbar_pattern(data: OHLC, order_type: OrderType) -> bool:
  """Pin-Bar Pattern.

  docs/images/buy_pinbar.png
  docs/images/sell_pinbar.png
  """
  opens = data.open
  highs = data.high
  lows = data.low
  closes = data.close
  if order_type.buy:
    upper_third = highs[-1] - (highs[-1] - lows[-1]) / 3
    body = abs(opens[-1] - closes[-1])
    lower_wick = opens[-1] - lows[-1]
    return (
        opens[-1] >= upper_third
    ) and (
        closes[-1] > upper_third
    ) and (
        lower_wick >= body * 3
    )
  else:
    lower_third = lows[-1] + (highs[-1] - lows[-1]) / 3
    body = abs(opens[-1] - closes[-1])
    upper_wick = highs[-1] - closes[-1]
    return (
        opens[-1] <= lower_third
    ) and (
        closes[-1] < lower_third
    ) and (
        upper_wick >= body * 3
    )


def harami_pattern(data: OHLC, order_type: OrderType) -> bool:
  """Harami Pattern.

  docs/images/buy_harami.png
  docs/images/sell_harami.png
  """
  opens = data.open
  highs = data.high
  lows = data.low
  closes = data.close
  if order_type.buy:
    return opens[-2] > closes[-2] and opens[-1] < closes[-1] and \
        highs[-1] < highs[-2] and lows[-1] > lows[-2]
  else:
    return opens[-2] < closes[-2] and opens[-1] > closes[-1] and \
        highs[-1] < highs[-2] and lows[-1] > lows[-2]



"""Parent class for event handlers."""
from __future__ import annotations
from typing import List, Dict, TYPE_CHECKING
from abc import ABC

if TYPE_CHECKING:
  from tradeo.order import Order
  from tradeo.ohlc import OHLC
  from tradeo.mt_client import MT_Client


class EventHandler(ABC):
  """This class should not be instantiated."""

  def __init__(self, event_handler_name: str):
    """Initialize the attributes."""
    self.event_handler_name = event_handler_name

  def on_tick(
      self,
      mt_client: MT_Client,
      symbol: str,
      bid: float,
      ask: float
  ) -> None:
    """Handle the return of SUBSCRIBE_SYMBOLS command."""
    return None

  def on_bar_data(
      self,
      mt_client: MT_Client,
      symbol: str,
      time_frame: str,
      time: str,
      ohlc: OHLC,
      tick_volume: float
  ) -> None:
    """Handle the return of SUBSCRIBE_SYMBOLS_BAR_DATA command."""
    return None

  def on_historical_data(
          self,
          mt_client: MT_Client,
          symbol: str,
          data: OHLC
  ) -> None:
    """Handle the return of GET_HISTORICAL_DATA command."""
    return None

  def on_historical_trades(
      self,
      mt_client: MT_Client
  ) -> None:
    """Handle the return of GET_HISTORICAL_TRADES command."""
    return None

  def on_message(
          self,
          mt_client: MT_Client,
          message: List[str]) -> None:
    """Handle when a new message is received."""
    return None

  def on_order_event(
          self,
          mt_client: MT_Client,
          account_info: Dict,
          open_orders: List[Order]
  ) -> None:
    """Handle when a new order event or removed order is received."""
    return None



"""Singleton definition."""


class Singleton(type):
  """Singleton metaclass."""

  _instances = {}

  def __call__(cls, *args, **kwargs):
    """Prevent multiple instances."""
    if cls not in cls._instances:
      cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
    return cls._instances[cls]



"""Script to handle logging."""

import logging
import socket
from logging.handlers import SysLogHandler
from logging import Handler
import requests

from tradeo.singleton import Singleton
from tradeo.config import Config


"""URL: https://my.papertrailapp.com/events ."""


class TelegramHandler(Handler):
  """Handler to send logs to telegram."""

  def emit(self, record):
    """Send the log to telegram."""
    log_entry = self.format(record)
    payload = {
        'chat_id': Config.forex_telegram_chat_id,
        'text': log_entry,
        'parse_mode': 'HTML'
    }
    token = Config.forex_telegram_token
    return requests.post(
        f'https://api.telegram.org/bot{token}/sendMessage',
        data=payload
    ).content


class ContextFilter(logging.Filter):
  """Filter to add hostname to log record."""

  hostname = socket.gethostname()

  def filter(self, record):  # noqa
    """Add the hostname 'e.g. raspberrypi' to the log record."""
    record.hostname = ContextFilter.hostname
    return True  # No filtering


class Log(metaclass=Singleton):
  """Class to manage logging."""

  def __init__(self):
    """Initialize the logger."""
    self.logger = logging.getLogger('tradeo')
    self.logger.setLevel(logging.DEBUG)
    self._build_log()
    self._build_telegram_log()

  def _build_log(self) -> None:
    handler = SysLogHandler(
        address=(Config.syslog_address, Config.syslog_port))

    # Filter
    handler.addFilter(ContextFilter())
    formatter = logging.Formatter(
        '%(asctime)s %(levelname)s trading-bot: %(message)s',
        datefmt='%b %d %H:%M:%S')

    # Format
    handler.setFormatter(formatter)

    # Level
    handler.setLevel(Config.log_level)

    # Add the handler
    self.logger.addHandler(handler)

  def _build_telegram_log(self) -> None:
    handler = TelegramHandler()

    # Level
    handler.setLevel(Config.tg_log_level)

    # Add the handler
    self.logger.addHandler(handler)

  def debug(self, msg: str):
    """Log debug message."""
    self.logger.debug(msg)

  def info(self, msg: str):
    """Log info message."""
    self.logger.info(msg)

  def warning(self, msg: str):
    """Log warning message."""
    self.logger.warning(f'🟡 {msg}')

  def error(self, msg: str):
    """Log error message."""
    msg = msg.replace('\n', ' ')
    self.logger.error(f'🔴 {msg}')


# Singleton instance
log = Log()



from tradeo.event_handlers.event_handler import EventHandler
from tradeo.event_handlers.basic_event_handler import BasicEventHandler

from tradeo.executable.executable import Executable
from tradeo.executable.basic_forex import BasicForex

from tradeo.strategies.strategy import Strategy
from tradeo.strategies.basic_strategy import BasicStrategy

from tradeo.config import Config
from tradeo.log import log
from tradeo.mt_client import MT_Client
from tradeo.utils import (
    string_to_date_utc, create_magic_number, get_last_balance,
    reboot_mt, reset_consecutive_times_down, increment_consecutive_times_down,
    get_consecutive_times_down
)
from tradeo.ohlc import OHLC
from tradeo.order_operations import OrderOperations
from tradeo.order_type import OrderType
from tradeo.order import (
    Order, MutableOrderDetails, ImmutableOrderDetails, OrderPrice
)
from tradeo.context_managers.blocker import Blocker
from tradeo.trading_methods import (get_pip, get_pivots, EMA, RSI, SAR,
                                    confirmation_pattern,
                                    three_bar_reversal, pinbar_pattern,
                                    harami_pattern
                                    )



"""Script to manage envs files."""
from pathlib import Path
import logging
import os
import pytz
from typing import Union


def _get_bool_from_env(env_name: str) -> Union[bool, None]:
  """Get bool from env."""
  e = os.getenv(env_name)
  if e is None:
    return None
  else:
    return str(e).lower() in ['true', '1']


def _get_logging_level(level: str) -> int:
  if level.lower() in ['info', 'inf', 'information']:
    log_level = logging.INFO
  elif level.lower() in ['debug', 'deb', 'debugging']:
    log_level = logging.DEBUG
  elif level.lower() in ['warning', 'warn']:
    log_level = logging.WARNING
  elif level.lower() in ['error', 'err']:
    log_level = logging.ERROR
  else:
    raise ValueError(f'Invalid log level: {level}')
  return log_level


class Config:
  """Class to manage config."""

  # Timezone Configuration
  local_timezone = pytz.timezone(
      os.getenv('TB_LOCAL_TIMEZONE') or 'Europe/Madrid')
  broker_timezone = pytz.timezone(
      os.getenv('TB_BROKER_TIMEZONE') or 'Etc/GMT-2')
  utc_timezone = pytz.utc

  # Paths configuration
  home = os.environ['HOME']
  default_mt_files_path = Path(
      f'{home}/.wine/drive_c/Program Files/MetaTrader/MQL5/Files'
  )
  mt_files_path = Path(os.getenv('TB_MT_FILES_PATH') or default_mt_files_path)

  # Trading configuration
  symbols = (os.getenv('TB_SYMBOLS') or 'EURUSD,USDJPY,USDCAD').split(',')
  account_currency = os.getenv('TB_ACCOUNT_CURRENCY') or 'EUR'
  timeframe = os.getenv('TB_TIMEFRAME') or 'M5'
  lookback_days = int(os.getenv('TB_LOOKBACK_DAYS') or 10)

  # Forex client configuration
  check_messages_thread = _get_bool_from_env(
      'TB_CHECK_MESSAGES_THREAD') or True
  check_market_data_thread = _get_bool_from_env(
      'TB_CHECK_MARKET_DATA_THREAD') or True
  check_bar_data_thread = _get_bool_from_env(
      'TB_CHECK_BAR_DATA_THREAD') or True
  check_open_orders_thread = _get_bool_from_env(
      'TB_CHECK_OPEN_ORDERS_THREAD') or True
  check_historical_data_thread = _get_bool_from_env(
      'TB_CHECK_HISTORICAL_DATA_THREAD') or True
  check_historical_trades_thread = _get_bool_from_env(
      'TB_CHECK_HISTORICAL_TRADES_THREAD') or True

  # Logging configuration
  syslog_address = os.getenv('TB_SYSLOG_ADDRESS') or 'logs2.papertrailapp.com'
  syslog_port = int(os.getenv('TB_SYSLOG_PORT') or 43931)
  log_level = _get_logging_level(os.getenv('TB_LOG_LEVEL') or 'DEBUG')

  # Telegram configuration
  forex_telegram_token = os.getenv(
      'TB_TG_FOREX_TOKEN') or ''
  forex_telegram_chat_id = int(os.getenv('TB_TG_FOREX_CHAT_ID') or 0)
  tg_log_level = _get_logging_level(os.getenv('TB_TG_LOG_LEVEL') or 'DEBUG')



"""This script is one of the possible entry points of the project."""
from datetime import datetime, timedelta
from random import randrange
import traceback

from tradeo.mt_client import MT_Client
from tradeo.files import write_file
from tradeo.files import Files
from tradeo.config import Config
from tradeo.utils import reboot_mt
from tradeo.log import log
from tradeo.executable.executable import Executable
from tradeo.strategies.basic_strategy import BasicStrategy
from tradeo.event_handlers.basic_event_handler import BasicEventHandler
from tradeo.context_managers.blocker import Blocker
from tradeo.utils import (reset_consecutive_times_down,
                          increment_consecutive_times_down,
                          get_consecutive_times_down, get_last_balance)


class BasicForex(Executable):
  """Basic example of a forex bot."""

  def __init__(self):
    """Initialize the forex bot."""
    self.name = 'BasicForex'

  def entry_point(self) -> None:
    """Entry point of the forex bot."""
    if not self.is_locked() and self.check_time_viability():
      mt_client = MT_Client(event_handler=BasicEventHandler())
      try:
        # Lock the execution of the forex bot.
        # Another thread can not run at the same time.
        with Blocker(name=self.name):
          self.main(mt_client)
      except Exception:  # noqa
        # Finish the bot
        self.finish(mt_client)
        # Log the error
        log.error(traceback.format_exc())

  def main(self, mt_client: MT_Client) -> None:
    """Execute forex bot."""
    # First of all, we lock the execution of the forex bot.
    # To prevent another execution to run at the same time.

    # Start the MT Client
    mt_client.start()

    # Clean all files
    mt_client.clean_all_command_files()
    mt_client.clean_all_historical_files()
    mt_client.clean_messages()

    # Dates
    utc_date = datetime.now(Config.utc_timezone)
    local_date = datetime.now(Config.local_timezone)

    # The instant of time that executed this main
    execution_time = timedelta(
        minutes=utc_date.minute % 5,
        seconds=utc_date.second,
        microseconds=utc_date.microsecond
    )

    # Send profit message
    self._send_profit_message(mt_client, local_date)

    # Send commands to obtain the historical data
    [
        mt_client.get_historical_data(s, Config.timeframe)
        for s in Config.symbols
    ]

    # Send commands to obtain bid/ask
    mt_client.subscribe_symbols(Config.symbols)

    # Handle the existing trades
    self.handle_trades(mt_client)

    # Process the result of "get_historical_data"
    self.handle_new_historical_data(mt_client, utc_date, execution_time)

    # Finish the main
    self.finish(mt_client)

  def handle_trades(self, mt_client: MT_Client) -> None:
    """Handle the existing trades."""
    orders = mt_client.check_open_orders()
    strategy = BasicStrategy()
    for order in orders:
      if order.order_type.pending:
        strategy.handle_pending_orders(mt_client, order)
      elif order.order_type.market:
        strategy.handle_filled_orders(mt_client, order)

    len_orders = len(orders)
    message = f'Number of open orders: {len_orders}'
    if len_orders > 900:
      log.warning(message)
    else:
      log.debug(message)

  def handle_new_historical_data(
          self,
          mt_client: MT_Client,
          utc_date: datetime,
          execution_time: timedelta
  ) -> None:
    """Handle the new historical data."""
    # Initialize the remaining symbols
    rs = mt_client.get_remaining_symbols()

    # The execution will take up to 4 minutes
    stop_condition = utc_date - execution_time + timedelta(minutes=4)

    while len(rs) > 0 and datetime.now(Config.utc_timezone) < stop_condition:
      # Get randomly the next symbol
      next_symbol = rs[randrange(len(rs))]

      # Check if JSON data is available to trigger the event
      mt_client.check_historical_data(next_symbol)

      # Update the remaining symbols
      rs = mt_client.get_remaining_symbols()

    # Check if there are remaining symbols to process
    if len(rs) > 0:
      log.warning(f'{len(rs)} remaining symbols to process.')
    else:
      reset_consecutive_times_down()

    # Check if MT needs to restart
    self._check_mt_needs_to_restart(len(rs))

  def _send_profit_message(
          self, mt_client: MT_Client, local_date: datetime) -> bool:
    """Get the balance of the account."""
    balance = mt_client.get_balance()
    last_balance = get_last_balance()
    difference = balance - last_balance
    emoji = '🚀' if difference >= 0 else '☔'
    message_condition = local_date.hour % 12 == 0 and local_date.minute == 5
    if message_condition and balance != -1:
      log.info(f'{emoji} {difference:.2f} €')
      write_file(Files.LAST_BALANCE.value, str(balance))
      return True
    else:
      return False

  def _check_mt_needs_to_restart(self, n_remaining_symbols: int) -> None:
    """Check if MT needs to restart."""
    ctd = get_consecutive_times_down()
    symbols_len = len(Config.symbols)
    if n_remaining_symbols > int(symbols_len / 2) and ctd > 4:
      reboot_mt()
    else:
      increment_consecutive_times_down()

  def is_locked(self) -> bool:
    """Return True if the forex-bot is running."""
    return Blocker(name=self.name).is_blocked()

  def check_time_viability(self) -> bool:
    """Check if the forex bot is viable to run."""
    now_date = datetime.now(Config.broker_timezone)
    # Monday (0) -> Sunday (6)
    is_weekday = now_date.weekday() in [0, 1, 2, 3, 4]
    # TODO: When executions are performed with the real account,
    # we need to consider testing the removal of this condition.
    is_not_on_the_hour = now_date.minute != 0
    return is_weekday and is_not_on_the_hour

  def finish(self, mt_client: MT_Client) -> None:
    """Finish the forex bot."""
    mt_client.stop()
    mt_client.deactivate()



"""Abstract class for strategies."""
from __future__ import annotations
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Union, TYPE_CHECKING

from tradeo.config import Config
from tradeo.log import log
if TYPE_CHECKING:
  from tradeo.order import Order
  from tradeo.ohlc import OHLC
  from tradeo.mt_client import MT_Client


class Strategy(ABC):
  """This class should not be instantiated."""

  def __init__(self, strategy_name: str):
    """Initialize the attributes."""
    self.strategy_name = strategy_name

  @abstractmethod
  def indicator(
          self,
          ohlc: OHLC,
          symbol: str,
          date: datetime
  ) -> Union[Order, None]:
    """Return an order if the strategy is triggered."""

  @staticmethod
  def check_order_viability(
          mt_client: MT_Client,
          order: Order,
          min_risk_profit: float = 1.5,
          **kwargs
  ) -> bool:
    """Check if the order is viable."""
    _ = kwargs
    symbol = order.symbol
    orders = [o for o in mt_client.open_orders if o.symbol == symbol]
    c1 = len(orders) == 0
    c2 = order.risk_benefit() > min_risk_profit
    # High spread
    c3 = datetime.now(Config.utc_timezone).hour not in [22, 23, 0]
    return c1 and c2 and c3

  @staticmethod
  def handle_pending_orders(
          mt_client: MT_Client,
          order: Order,
          time_threshold: int = 3600
  ) -> None:
    """Handle limit orders based on the time threshold."""
    try:
      open_time = datetime.fromtimestamp(
          int(order.magic)).astimezone(Config.utc_timezone)
      current_datetime = datetime.now(Config.utc_timezone)
      if (current_datetime - open_time).seconds > time_threshold:
        mt_client.send_close_orders_by_magic_command(order.magic)
        log.debug(f'Close order {order.magic} due to time threshold')
    except ValueError:  # int(order.magic)
      pass

  @staticmethod
  def handle_filled_orders(
      mt_client: MT_Client,
      order: Order,
      time_threshold: int = 3600 * 24,
      break_even_time_threshold: int = 3600 * 12,
      break_even_per_threshold: float = 0.75,
      **kwargs
  ) -> None:
    """Handle filled orders by closing them or placing a break even."""
    _ = kwargs
    try:
      open_time = datetime.fromtimestamp(
          int(order.magic)).astimezone(Config.utc_timezone)
      current_datetime = datetime.now(Config.utc_timezone)

      # Check if the order can be closed based on the time threshold
      if (current_datetime - open_time).seconds > time_threshold:
        mt_client.send_close_orders_by_magic_command(order.magic)
        log.debug(f'Close order {order.magic} due to time threshold')

      # Check if a break even can be placed
      else:
        Strategy._check_if_break_even_can_be_placed(
            mt_client,
            order,
            open_time,
            current_datetime,
            break_even_time_threshold,
            break_even_per_threshold
        )

    except ValueError:  # int(order.magic)
      pass

  @staticmethod
  def _check_if_break_even_can_be_placed(
      mt_client: MT_Client,
      order: Order,
      open_time: datetime,
      current_datetime: datetime,
      break_even_time_threshold: int,
      break_even_per_threshold: float,
      **kwargs
  ) -> bool:
    """Check if a break even can be placed."""
    _ = kwargs
    result = False
    # First check if a previous break even has been placed
    break_even_placed_buy = (
        order.order_type.buy and order.stop_loss > order.price
    )
    break_even_placed_sell = (
        order.order_type.sell and order.stop_loss < order.price
    )
    break_even_placed = break_even_placed_buy or break_even_placed_sell

    # Check if the time has reached a threshold to place a break even
    if not break_even_placed:
      reached_even_time_threshold = (
          current_datetime - open_time
      ).seconds > break_even_time_threshold

    # Check if the price has reached a threshold to place a break even
    if not break_even_placed:
      bid, ask = mt_client.get_bid_ask(order.symbol)
      price = (bid + ask) / 2
      percentage_reached = (
          price - order.stop_loss
      ) / (
          order.take_profit - order.stop_loss
      )
      price_reached_threshold = percentage_reached >= break_even_per_threshold

    if not break_even_placed and (
        price_reached_threshold or reached_even_time_threshold
    ):
      # We place a break even
      mt_client.place_break_even(order)
      result = True

    return result



"""Script to collect different files utilities."""
import typing as ty
from pathlib import Path
from os.path import exists
import json
import os
from time import sleep
from json.decoder import JSONDecodeError
from enum import Enum

from tradeo.paths import get_default_path

_default_path = get_default_path()


class Files(str, Enum):
  """Files enum."""

  # GENERAL
  SUCCESSFUL_SYMBOLS = 'successful_symbols.txt'
  CONSECUTIVE_TIMES_DOWN = 'consecutive_times_down.txt'
  LAST_BALANCE = 'last_balance.txt'


def file_exists(file: str, file_path: Path = _default_path) -> bool:
  """Return True if the file exists.

  file_path: "get_default_path()" as default value.
  """
  path = file_path if file_path != _default_path else get_default_path()
  return exists(path / file)


def try_load_json(file_path: Path) -> ty.Dict[str, ty.Dict]:
  """Try to load a JSON from a file generate from MQL."""
  for _ in range(5):
    try:
      if exists(file_path):
        with open(file_path, 'r') as f:
          text = f.read()
          return json.loads(text)
    except (IOError, JSONDecodeError):
      pass
    sleep(0.1)
  return {}


def try_read_file(file_path: Path) -> str:
  """Try to read a file."""
  try:
    with open(file_path, 'r') as f:
      return f.read()
  except IOError:
    pass
  return ''


def write_file(
    file_name: str,
    text: str = '',
    mode: str = 'w',
    file_path: Path = _default_path
) -> None:
  """Write a file.

  file_path: "get_default_path()" as default value.
  """
  path = file_path if file_path != _default_path else get_default_path()

  with open(file=path / file_name, mode=mode) as f:
    f.write(str(text))


def remove_file(
    file_name: str,
    file_path: Path = _default_path
) -> None:
  """Remove a file."""
  path = file_path if file_path != _default_path else get_default_path()
  os.remove(path / file_name)


def try_remove_file(file_path: Path) -> bool:
  """Try to remove a file."""
  for _ in range(5):
    if file_path.exists():
      try:
        os.remove(file_path)
        return True
      except IOError:
        pass
      sleep(0.1)
  return False



"""Class than extends EventHandler to provide basic functionality."""
from __future__ import annotations
from datetime import datetime
from typing import TYPE_CHECKING

from tradeo.event_handlers.event_handler import EventHandler
from tradeo.strategies.basic_strategy import BasicStrategy
from tradeo.config import Config
if TYPE_CHECKING:
  from tradeo.ohlc import OHLC
  from tradeo.mt_client import MT_Client


class BasicEventHandler(EventHandler):
  """This class only provides 'on_historical_data' event holder."""

  def __init__(self):
    """Initialize the attributes."""
    super().__init__('BasicEventHandler')

  def on_historical_data(
          self,
          mt_client: MT_Client,
          symbol: str,
          data: OHLC
  ) -> None:
    """Handle the return of GET_HISTORICAL_DATA command."""
    now_date = datetime.now(Config.utc_timezone)
    strategy = BasicStrategy()
    possible_order = strategy.indicator(data, symbol, now_date)
    if possible_order and strategy.check_order_viability(
            mt_client, possible_order):
      mt_client.create_new_order(possible_order)



"""Script of MT_Client what it sends commands to MT4/MT5."""
from __future__ import annotations
from datetime import datetime, timedelta
from typing import List, Dict, Union, Callable, Tuple, TYPE_CHECKING, Set, cast
from threading import Thread, Lock
from os.path import join, exists
from random import randrange
from pandas import DataFrame
from pathlib import Path
from time import sleep
import glob
import json

from tradeo.config import Config
from tradeo.log import log
from tradeo.singleton import Singleton
from tradeo.files import (
    try_load_json, try_remove_file, try_read_file)
from tradeo.order_operations import OrderOperations
from tradeo.order_type import get_order_type_from_str
from tradeo.order import (
    Order, MutableOrderDetails, ImmutableOrderDetails, OrderPrice)
from tradeo.ohlc import OHLC
from tradeo.trading_methods import get_pip
from tradeo.utils import string_to_date_utc
from tradeo.mt_message import MT_MessageError, MT_MessageInfo
if TYPE_CHECKING:
  from tradeo.event_handlers.event_handler import (
      EventHandler)  # pragma: no cover


# Typing types
attributes_data_type = Dict[str, Dict]
messages_type = Dict[str, List[Union[MT_MessageInfo, MT_MessageError]]]
account_info_type = Dict[str, Union[float, str]]
historical_data_type = Dict[str, DataFrame]


class MT_Client(metaclass=Singleton):
  """Forex Client MetaTrader class.

  This includes all of the functions needed for communication with MT4/MT5.
  """

  def __init__(self,
               event_handler: Union[EventHandler, None] = None,
               sleep_delay: float = 0.005,
               max_retry_command_seconds: int = 10,
               files_subfolder: str = 'AgentFiles'
               ):
    """Initialize the attributes."""
    # Parameter attributes
    self.sleep_delay = sleep_delay
    self.max_retry_command_seconds = max_retry_command_seconds
    self.num_command_files = 50

    # Paths to output MT files
    self.prefix_files_path = files_subfolder
    self.set_agent_paths()

    # Control attributes
    self.event_handler = event_handler
    self.lock = Lock()
    self._last_messages_millis = 0
    self.command_id = 0

    # Data attributes
    self.messages: messages_type = {'INFO': [], 'ERROR': []}
    self.open_orders: List[Order] = []
    self.account_info: account_info_type = {}
    self.market_data: attributes_data_type = {}
    self.bar_data: attributes_data_type = {}
    self.historical_data: historical_data_type = {}
    self.historical_trades: attributes_data_type = {}
    self._successful_symbols: Set[str] = set()

    # State attributes
    self.activate()

  def set_agent_paths(self) -> None:
    """Set the paths to the files generated by MQL."""
    mt_files_path = Config.mt_files_path
    if Path(mt_files_path).exists():
      prefix_folder = Path(mt_files_path / self.prefix_files_path)
      prefix_folder.mkdir(exist_ok=True)
      self.path_orders = Path(
          join(mt_files_path, self.prefix_files_path, 'Orders.json'))
      self.path_messages = Path(
          join(mt_files_path, self.prefix_files_path, 'Messages.json'))
      self.path_market_data = Path(
          join(mt_files_path, self.prefix_files_path, 'Market_Data.json'))
      self.path_bar_data = Path(
          join(mt_files_path, self.prefix_files_path, 'Bar_Data.json'))
      self.path_historical_data_prefix = Path(
          join(mt_files_path, self.prefix_files_path, 'Historical_Data_'))
      self.path_historical_trades = Path(
          join(mt_files_path, self.prefix_files_path,
               'Historical_Trades.json'))
      self.path_orders_stored = Path(
          join(mt_files_path, self.prefix_files_path,
               'Orders_Stored.json'))
      self.path_messages_stored = Path(
          join(mt_files_path, self.prefix_files_path,
               'Messages_Stored.json'))
      self.path_commands_prefix = Path(
          join(mt_files_path, self.prefix_files_path, 'Commands_'))
    else:
      log.error(f'mt_files_path: {mt_files_path} does not exist!')

  @property
  def successful_symbols(self) -> Set[str]:
    """Return the set of successful symbols."""
    return self._successful_symbols

  @staticmethod
  def start_thread(target: Callable) -> Thread:
    """To start the thread with a method as target."""
    thread = Thread(target=target, args=(), daemon=True)
    thread.start()
    return thread

  def start(self) -> None:
    """Start the threads."""
    self.START = True
    self.send_reset_command_ids_command()
    # Start the demand threads
    if Config.check_messages_thread:
      self.start_thread(self.start_thread_check_messages)
    if Config.check_market_data_thread:
      self.start_thread(self.start_thread_check_market_data)
    if Config.check_bar_data_thread:
      self.start_thread(self.start_thread_check_bar_data)
    if Config.check_open_orders_thread:
      self.start_thread(self.start_thread_check_open_orders)
    if Config.check_historical_data_thread:
      self.start_thread(self.start_thread_check_historical_data)
    if Config.check_historical_trades_thread:
      self.start_thread(self.start_thread_check_historical_trades)

  def stop(self) -> None:
    """Stop the threads."""
    self.START = False

  def activate(self) -> None:
    """Activate the threads."""
    self.ACTIVE = True

  def deactivate(self) -> None:
    """Deactivate the threads."""
    self.ACTIVE = False

  def start_thread_check_messages(self) -> None:
    """Start the thread to check messages."""
    while self.ACTIVE:
      sleep(self.sleep_delay)
      if self.START:
        self.check_messages()

  def check_messages(self) -> messages_type:
    """Update and return the messages object."""
    data = try_load_json(self.path_messages)

    if len(data) > 0 and data != self.messages:

      for millis, message in sorted(data.items()):
        if int(millis) > self._last_messages_millis:
          self._last_messages_millis = int(millis)

          message_content = list(message.values())
          if 'ERROR' in message_content:
            time = message_content[1]
            error_type = message_content[2]
            description = message_content[3]
            self.messages['ERROR'].append(
                MT_MessageError(time, error_type, description)
            )
          else:
            time = message_content[1]
            message = message_content[2]
            self.messages['INFO'].append(
                MT_MessageInfo(time, message)
            )

          if self.event_handler:
            self.event_handler.on_message(self, message_content)

    return self.messages

  def get_error_messages(self) -> List[MT_MessageError]:
    """Return the error messages."""
    return cast(List[MT_MessageError], self.messages['ERROR'])

  def get_info_messages(self) -> List[MT_MessageInfo]:
    """Return the info messages."""
    return cast(List[MT_MessageInfo], self.messages['INFO'])

  def set_messages(
      self,
      info_messages: List[MT_MessageInfo],
      error_messages: List[MT_MessageError]
  ) -> None:
    """Set manually the messages object."""
    self.messages['INFO'] = cast(
        List[Union[MT_MessageInfo, MT_MessageError]], info_messages
    )
    self.messages['ERROR'] = cast(
        List[Union[MT_MessageInfo, MT_MessageError]], error_messages
    )

  def clean_messages(self):
    """Clean the messages object."""
    try_remove_file(self.path_messages)
    self.messages = {'INFO': [], 'ERROR': []}

  def start_thread_check_market_data(self) -> None:
    """Start the thread to check market data."""
    while self.ACTIVE:
      sleep(self.sleep_delay)
      if self.START:
        self.check_market_data()

  def check_market_data(self) -> Dict[str, Dict]:
    """Update, trigger event if needed and return the market data object."""
    data = try_load_json(self.path_market_data)

    if len(data) > 0 and data != self.market_data:

      if self.event_handler:
        for symbol in data.keys():
          cond = symbol not in self.market_data
          cond = cond or data[symbol] != self.market_data.get(symbol)
          if cond:
            self.event_handler.on_tick(
                self,
                symbol,
                data[symbol]['bid'],
                data[symbol]['ask']
            )
      self.market_data = data

    return self.market_data

  def get_bid_ask(self, symbol: str) -> Tuple[float, float]:
    """Return the bid and ask price of a symbol.

    Return 0,0 if symbol not found.

    Bid: sell price
    Ask: buy price
    """
    try:
      bid_ask = self.market_data[symbol]
      return bid_ask['bid'], bid_ask['ask']
    except KeyError:
      log.warning(f'Symbol {symbol} not found in market data.')
      return 0, 0

  def start_thread_check_bar_data(self) -> None:
    """Start the thread to check bar data."""
    while self.ACTIVE:
      sleep(self.sleep_delay)
      if self.START:
        self.check_bar_data()

  def check_bar_data(self) -> Dict[str, Dict]:
    """Update, trigger event if needed and return the bar data object."""
    data = try_load_json(self.path_bar_data)

    if len(data) > 0 and data != self.bar_data:

      if self.event_handler:
        for st in data.keys():
          cond = st not in self.bar_data
          cond = cond or data[st] != self.bar_data[st]
          if cond:
            symbol, time_frame = st.split('_')
            self.event_handler.on_bar_data(
                self,
                symbol,
                time_frame,
                data[st]['time'],
                OHLC(DataFrame({
                    'open': data[st]['open'],
                    'high': data[st]['high'],
                    'low': data[st]['low'],
                    'close': data[st]['close'],
                })),
                data[st]['tick_volume']
            )
      self.bar_data = data

    return self.bar_data

  def start_thread_check_open_orders(self) -> None:
    """Start the thread to check open orders."""
    while self.ACTIVE:
      sleep(self.sleep_delay)
      if self.START:
        self.check_open_orders()

  def check_open_orders(self) -> List[Order]:
    """Update, trigger event if needed and return the open orders object.

    The open orders can be pending or filled.
    """
    data = try_load_json(self.path_orders)
    data_orders = data.get('orders')
    data_account_info = data.get('account_info')

    if (len(data) > 0
        and isinstance(data_orders, Dict)
        and isinstance(data_account_info, Dict)
        and (data_orders != self.open_orders
             or data_account_info != self.account_info)):

      orders = self._transform_json_orders_to_orders(data_orders)

      new_event = False
      # If an existing open order is not in the new data, trigger an event
      for order in self.open_orders:
        if order not in orders:
          new_event = True
          break

      # If a new open order is not in the existing data, trigger an event
      for order in orders:
        if order not in self.open_orders:
          new_event = True
          break

      self.account_info = data_account_info
      self.open_orders = orders

      with open(self.path_orders_stored, 'w') as f:
        f.write(json.dumps(data))

      if new_event and self.event_handler:
        self.event_handler.on_order_event(
            self, self.account_info, self.open_orders
        )

    return self.open_orders

  def _transform_json_orders_to_orders(
          self, json_orders: Dict) -> List[Order]:
    """Return a list of open Order objects."""
    return [
        Order(
            MutableOrderDetails(
                OrderPrice(
                    price=o['open_price'],
                    stop_loss=o['SL'],
                    take_profit=o['TP']
                ), lots=o['lots']
            ),
            ImmutableOrderDetails(
                symbol=o['symbol'],
                order_type=get_order_type_from_str(o['type']),
                magic=o['magic'],
                comment=o['comment']
            ), ticket=int(t), pnl=o['pnl'])
        for t, o in json_orders.items()
    ]

  def start_thread_check_historical_data(self) -> None:
    """Start the thread to check historical data."""
    while self.ACTIVE:
      sleep(self.sleep_delay)
      if self.START:
        self.check_historical_data()

  def check_historical_data(
          self, symbol: Union[str, None] = None) -> Dict:
    """Update historical_data, trigger event if needed and return that data."""
    # "symbol" is None when it comes from "start_thread_check_historical_data"
    # In this case, we need to get a random remaining symbol
    remaining_symbols = self.get_remaining_symbols()
    if len(remaining_symbols) == 0:
      return {}
    if symbol is None:
      symbol = remaining_symbols[randrange(len(remaining_symbols))]

    # We read the symbol file
    file_path = Path(f'{self.path_historical_data_prefix}{symbol}.json')
    data = try_load_json(file_path)

    if len(data) > 0:
      # The dataframe is built
      df = DataFrame.from_dict(
          data[f'{symbol}_{Config.timeframe}'], orient='index')
      self.historical_data[symbol] = df

      # The date and time corresponding to the last load are calculated
      if self._is_historical_data_up_to_date(df):
        log.debug(f'{symbol} -> {(df.index[0], df.index[-1])}')

        self.successful_symbols.add(symbol)
        if self.event_handler:
          self.event_handler.on_historical_data(self, symbol, OHLC(df))

        # We delete the command file(s) in case it hasn't been deleted.
        command_files = self.command_file_exist(symbol)
        for com in command_files:
          try_remove_file(com)

    return data

  @staticmethod
  def _is_historical_data_up_to_date(df: DataFrame) -> bool:
    """Check if the historical data is up to date."""
    now_date = datetime.now(Config.utc_timezone)
    td = timedelta(minutes=now_date.minute % 5,
                   seconds=now_date.second,
                   microseconds=now_date.microsecond)
    rounded_now_date = now_date - td
    start_range = rounded_now_date
    end_range = rounded_now_date + timedelta(minutes=5)
    last_date_from_df = string_to_date_utc(
        str_date=df.index[-1], from_timezone=Config.broker_timezone)

    return (last_date_from_df >= start_range
            and last_date_from_df < end_range)

  def start_thread_check_historical_trades(self) -> None:
    """Start the thread to check historical trades."""
    while self.ACTIVE:
      sleep(self.sleep_delay)
      if self.START:
        self.check_historical_trades()

  def check_historical_trades(self) -> Dict:
    """Update and return the historical trades object."""
    self.historical_trades = try_load_json(self.path_historical_trades)
    return self.historical_trades

  def subscribe_symbols(self, symbols: List[str]) -> None:
    """To send a SUBSCRIBE_SYMBOLS command to subscribe to market (tick) data.

    Args:
        symbols (list[str]): List of symbols to subscribe to.

    Returns:
        None

        The data will be stored in self.market_data.
        On receiving the data the event_handler.on_tick()
        function will be triggered.

    """
    self.send_command('SUBSCRIBE_SYMBOLS', ','.join(symbols))

  def subscribe_symbols_bar_data(
      self,
      symbols: List[List[str]]
  ) -> None:
    """To send a SUBSCRIBE_SYMBOLS_BAR_DATA command to subscribe to bar data.

    Kwargs:
        symbols (list[list[str]]): List of lists containing symbol/time frame
        combinations to subscribe to. For example:
        symbols = [['EURUSD', 'M1'], ['GBPUSD', 'H1']]

    Returns:
        None

        The data will be stored in self.bar_data.
        On receiving the data the event_handler.on_bar_data()
        function will be triggered.

    """
    data = [f'{st[0]},{st[1]}' for st in symbols]
    self.send_command('SUBSCRIBE_SYMBOLS_BAR_DATA',
                      ','.join(str(p) for p in data))

  def get_historical_data(
      self,
      symbol: str,
      time_frame: str
  ) -> None:
    """To send a GET_HISTORIC_DATA command to request historical data.

    Kwargs:
        symbol (str): Symbol to get historical data.
        time_frame (str): Time frame for the requested data.

    Returns:
        None

        The data will be stored in self.historical_data.
        On receiving the data the event_handler.on_historical_data()
        function will be triggered.
    """
    # We add 10 hours because the way the library interprets this input requires
    # overshooting to ensure capturing up to the last record.
    end = datetime.now(Config.broker_timezone) + timedelta(hours=10)
    start = (end - timedelta(days=Config.lookback_days)).timestamp()
    end = end.timestamp()
    data = [symbol, time_frame, int(start), int(end)]
    self.send_command('GET_HISTORICAL_DATA', ','.join(str(p) for p in data))

  def get_historical_trades(self, lookback_days: int = 30) -> None:
    """To send a GET_HISTORIC_TRADES command to request historical trades.

    Kwargs:
        lookback_days (int): Days to look back into the trade history.
        The history must also be visible in MT4.

    Returns:
        None

        The data will be stored in self.historical_trades.
        On receiving the data the event_handler.on_historical_trades()
        function will be triggered.
    """
    self.send_command('GET_HISTORICAL_TRADES', str(lookback_days))

  def create_new_order(self, order: Order) -> None:
    """Create new order."""
    log.debug(f'Creating new order: {order}')

    if order.order_type.pending:
      bid, ask = self.get_bid_ask(order.symbol)
      self._modify_pending_order(order, bid, ask)

    self.send_open_order_command(order)

  @staticmethod
  def _modify_pending_order(order: Order, bid: float, ask: float) -> None:
    """Modify the pending order based on the bid/ask."""
    if bid != 0 and ask != 0:
      ot = order.order_type
      if ot.buy and order.price > ask:
        ot.value = OrderOperations.BUYSTOP
      elif ot.buy and order.price < ask:
        ot.value = OrderOperations.BUYLIMIT
      elif ot.sell and order.price < bid:
        ot.value = OrderOperations.SELLSTOP
      elif ot.sell and order.price > bid:
        ot.value = OrderOperations.SELLLIMIT

  def send_open_order_command(self, order: Order) -> None:
    """To send an OPEN_ORDER command to open an order."""
    data = [
        order.symbol, order.order_type.value.value, order.lots, order.price,
        order.stop_loss, order.take_profit, order.magic, order.comment,
        order.expiration
    ]
    self.send_command('OPEN_ORDER', ','.join(str(p) for p in data))

  def send_modify_order_command(
      self,
      ticket: int,
      mod: MutableOrderDetails
  ) -> None:
    """To send a MODIFY_ORDER command to modify an order.

    Args:
        ticket (int): Ticket of the order that should be modified.

    Kwargs:
        lots (float): Volume in lots
        price (float): Price of the (pending) order. Non-zero only
            works for pending orders.
        stop_loss (float): New stop loss price.
        take_profit (float): New take profit price.
        expiration (int): New expiration time given as timestamp in seconds.
            Can be zero if the order should not have an expiration time.

    """
    data = [
        ticket, mod.lots, mod.price, mod.stop_loss, mod.take_profit,
        mod.expiration
    ]
    self.send_command('MODIFY_ORDER', ','.join(str(p) for p in data))

  def place_break_even(self, order: Order) -> None:
    """Modify the order to place a break even."""
    pip = get_pip(order.symbol)
    buy = order.order_type.buy
    break_even = order.price + pip if buy else order.price - pip
    self.send_modify_order_command(
        order.ticket,
        MutableOrderDetails(
            prices=OrderPrice(
                price=break_even,
                stop_loss=order.stop_loss,
                take_profit=order.take_profit
            ),
            lots=order.lots,
            expiration=order.expiration
        )
    )
    log.debug(f'Break even placed in {order.magic}')

  def send_close_order_command(self, ticket: int, lots: float = 0) -> None:
    """To send a CLOSE_ORDER command to close an order.

    Args:
        ticket (int): Ticket of the order that should be closed.

    Kwargs:
        lots (float): Volume in lots. If lots=0 it will try to
            close the complete position.

    """
    data = [ticket, lots]
    self.send_command('CLOSE_ORDER', ','.join(str(p) for p in data))

  def send_close_all_orders_command(self) -> None:
    """To send a CLOSE_ALL_ORDERS command to close all orders."""
    self.send_command('CLOSE_ALL_ORDERS', '')

  def send_close_orders_by_symbol_command(self, symbol: str) -> None:
    """To send a CLOSE_ORDERS_BY_SYMBOL command to close all orders.

    Args:
        symbol (str): Symbol for which all orders should be closed.

    """
    self.send_command('CLOSE_ORDERS_BY_SYMBOL', symbol)

  def send_close_orders_by_magic_command(self, magic: str) -> None:
    """To send a CLOSE_ORDERS_BY_MAGIC command to close all orders.

    Args:
        magic (str): Magic number for which all orders should
            be closed.

    """
    self.send_command('CLOSE_ORDERS_BY_MAGIC', magic)

  def send_reset_command_ids_command(self) -> None:
    """To send a RESET_COMMAND_IDS command to reset stored command IDs.

    This should be used when restarting the python side without restarting
    the mql side.
    """
    self.command_id = 0

    self.send_command('RESET_COMMAND_IDS', '')

    # sleep to make sure it is read before other commands.
    sleep(0.5)

  def send_command(self, command: str, content: str) -> None:
    """To send a command to the MQL side.

    Multiple command files are used to allow for fast execution
    of multiple commands in the correct chronological order.

    """
    # Acquire lock so that different threads do not use the same
    # command_id or write at the same time.
    self.lock.acquire()

    self.command_id = (self.command_id + 1) % 100000

    end_time = datetime.now(Config.utc_timezone) + timedelta(
        seconds=self.max_retry_command_seconds)
    now = datetime.now(Config.utc_timezone)

    # trying again for X seconds in case all files exist or are
    # currently read from mql side.
    while now < end_time:
      # using 10 different files to increase the execution speed
      # for multiple commands.
      success = False
      for i in range(self.num_command_files):
        # only send commend if the file does not exists so that we
        # do not overwrite all commands.
        file_path = f'{self.path_commands_prefix}{i}.txt'
        if not exists(file_path):
          with open(file_path, 'w') as f:
            f.write(f'<:{self.command_id}|{command}|{content}:>')
          success = True
          break
      if success:
        break
      sleep(self.sleep_delay)
      now = datetime.now(Config.utc_timezone)

    # release lock again
    self.lock.release()

  def clean_all_command_files(self) -> None:
    """Clean command files."""
    for i in range(self.num_command_files):
      file_path = Path(f'{self.path_commands_prefix}{i}.txt')
      if not try_remove_file(file_path):
        break

  def clean_all_historical_files(self) -> None:
    """Clean historical files."""
    for symbol in Config.symbols:
      file_path = Path(f'{self.path_historical_data_prefix}{symbol}.json')
      try_remove_file(file_path)

  def command_file_exist(self, symbol: str) -> List[Path]:
    """Return the command files that match request hist. data from symbol."""
    g = glob.glob(f'{self.path_commands_prefix}*')
    pattern = f'GET_HISTORICAL_DATA|{symbol}'
    return [
        Path(f) for f in g if pattern in try_read_file(Path(f))
    ]

  def get_balance(self) -> float:
    """Return the balance of the account.

    Returns -1.0 if the balance is not a float.
    """
    max_retries = 5
    for _ in range(max_retries):
      self.check_open_orders()
      balance = self.account_info.get('balance')
      if isinstance(balance, float):
        return balance
      sleep(0.5)
    log.warning(f'Balance is not a float: {balance}')
    return -1.0

  def get_remaining_symbols(self) -> List[str]:
    """Return the list of remaining symbols."""
    all_symbols = set(Config.symbols)
    return list(all_symbols - self.successful_symbols)

  def get_lot_size(
      self,
      symbol: str,
      entry_price: float,
      stop_loss: float,
      risk_ratio: float
  ) -> float:
    """Return the lot size based on the risk (0-100%) on your account.

    Returns 0.01 if the bid and ask are not available or if the risk is lower
    than 0.01.
    """
    bid, ask = self.get_bid_ask(Config.account_currency + symbol[3:6])
    if bid != 0 and ask != 0:
      price = (bid + ask) / 2
      pip = get_pip(symbol)
      sl_pips = abs(entry_price - stop_loss) / pip
      balance = self.get_balance()
      risk_ratio = risk_ratio / 100
      lots = round(
          (risk_ratio * balance) / ((100000 / price * pip) * sl_pips), 2
      )
      return lots if lots > 0 else 0.01
    else:
      return 0.01



"""Handle every type of trading operations."""
from enum import Enum


class OrderOperations(str, Enum):
  """Different order operations."""

  BUYLIMIT = 'buylimit'
  BUYSTOP = 'buystop'
  SELLLIMIT = 'selllimit'
  SELLSTOP = 'sellstop'
  BUY = 'buy'
  SELL = 'sell'



"""Abstract class for executables."""
from abc import ABC, abstractmethod
from datetime import datetime, timedelta

from tradeo.mt_client import MT_Client


class Executable(ABC):
  """This class should not be instantiated."""

  @abstractmethod
  def entry_point(self) -> None:
    """Entry point of the executable."""

  @abstractmethod
  def main(self, mt_client: MT_Client) -> None:
    """Execute the executable."""

  @abstractmethod
  def handle_trades(self, mt_client: MT_Client) -> None:
    """Handle the existing trades."""

  @abstractmethod
  def handle_new_historical_data(
          self,
          mt_client: MT_Client,
          utc_date: datetime,
          execution_time: timedelta
  ) -> None:
    """Handle the new historical data."""

  @abstractmethod
  def is_locked(self) -> bool:
    """Return True if the executable is running."""

  @abstractmethod
  def check_time_viability(self) -> bool:
    """Check if the time is viable to execute the executable."""

  @abstractmethod
  def finish(self, mt_client: MT_Client) -> None:
    """Finish the executable."""



"""MT message classes."""
from tradeo.utils import string_to_date_utc
from tradeo.config import Config


class MT_MessageError:
  """Class for handling errors in MT messages."""

  def __init__(self, time: str, error_type: str, description: str):
    """Initialize the attributes."""
    self.time = string_to_date_utc(
        time,
        date_format='%Y.%m.%d %H:%M:%S',
        from_timezone=Config.broker_timezone
    )
    self.error_type = error_type
    self.description = description

  def __eq__(self, value: object) -> bool:
    """Check if the object is equal to another object."""
    return (
        isinstance(value, MT_MessageError)
        and str(self.time.year) == str(value.time.year)
        and str(self.time.month) == str(value.time.month)
        and str(self.time.day) == str(value.time.day)
        and str(self.time.hour) == str(value.time.hour)
        and str(self.time.minute) == str(value.time.minute)
        and str(self.time.second) == str(value.time.second)
        and self.error_type == value.error_type
        and self.description == value.description
    )


class MT_MessageInfo:
  """Class for handling info messages in MT messages."""

  def __init__(self, time: str, message: str):
    """Initialize the attributes."""
    self.time = string_to_date_utc(
        time,
        date_format='%Y.%m.%d %H:%M:%S',
        from_timezone=Config.broker_timezone
    )
    self.message = message

  def __eq__(self, value: object) -> bool:
    """Check if the object is equal to another object."""
    return (
        isinstance(value, MT_MessageInfo)
        and str(self.time.year) == str(value.time.year)
        and str(self.time.month) == str(value.time.month)
        and str(self.time.day) == str(value.time.day)
        and str(self.time.hour) == str(value.time.hour)
        and str(self.time.minute) == str(value.time.minute)
        and str(self.time.second) == str(value.time.second)
        and self.message == value.message
    )



"""Script to collect different path utilities."""
from pathlib import Path


def config_path() -> Path:
  """Return the config path."""
  return root_project() / 'config'


def bash_path() -> Path:
  """Return the bash path."""
  return bin_project() / 'bash'


def resources_test_path() -> Path:
  """Return the test resources path."""
  return root_project() / 'tests' / 'resources'


def get_default_path():
  """Return the default path."""
  return data_path()


def data_path() -> Path:
  """Return the data path."""
  return bin_project() / 'data'


def root_project() -> Path:
  """Return the root project path."""
  return bin_project().parent


def bin_project() -> Path:
  """Return the root project path."""
  return Path(__file__).parent
